from scipy.stats import skew
from scipy.stats import kurtosis
from sklearn.metrics import accuracy_score
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score
from sklearn.metrics import f1_score
from sklearn.feature_selection import SelectKBest
from sklearn.feature_selection import f_classif
from sklearn.feature_selection import VarianceThreshold
from sklearn.model_selection import train_test_split
from sklearn.ensemble import RandomForestClassifier
import numpy as np
import time


def load_feature_values(filename):
    return np.load(filename)


def load_target_values(filename):
    return np.genfromtxt(filename, delimiter=",")


def split_dataset(x, y, test_size, random_state):
    return train_test_split(x, y, test_size=test_size, random_state=random_state)


def remove_zero_variance_features(x):
    x = x.reshape(x.shape[0], x.shape[1] * x.shape[2])
    x_new = VarianceThreshold().fit_transform(x)
    print("Removed features with no variance.")
    print("Number of features was reduced from " + str(x.shape[1]) + " to " + str(x_new.shape[1]) + ".")
    x = x_new
    return x


def select_k_best_features(k_best_features, x, y):
    print("Selecting K best features...")
    x = SelectKBest(f_classif, k=k_best_features).fit_transform(x, y)
    print(str(k_best_features) + " K best features selected.")
    return x


def get_fraction_of_dataset(fraction, x_train, y_train, x_test, y_test):
    print ("Extracting fraction of dataset...")
    x_train = x_train[:int(len(x_train) * fraction)]
    x_test = x_test[:int(len(x_test) * fraction)]
    y_train = y_train[:int(len(y_train) * fraction)]
    y_test = y_test[:int(len(y_test) * fraction)]
    print (str(fraction) + " of dataset extracted.")
    return x_train, y_train, x_test, y_test


def calculate_feature_imbalance(x):
    print ("Calculating feature imbalance...")
    print ("Kurtosis: " + str(kurtosis(x)))
    print ("Skewness: " + str(skew(x)))


def calculate_performance_scores(revised_performance_measures, y_test, y_pred):
    accuracy = accuracy_score(y_true=y_test, y_pred=y_pred, normalize=True)

    Tn = 0
    Tp = 0
    Fn = 0
    Fp = 0
    tot = len(y_pred)
    for i, pred in enumerate(y_pred):
        if pred == y_test[i]:
            if pred == 0:
                Tn += 1
            elif pred == 1:
                Tp += 1
        elif pred != y_test[i]:
            if pred == 0:
                Fn += 1
            elif pred == 1:
                Fp += 1

    Tn = Tn / tot
    Tp = Tp / tot
    Fn = Fn / tot
    Fp = Fp / tot

    if revised_performance_measures:
        precision = (Tn / (Tn + Fn))
        recall = (Tn / (Tn + Fp))
        f1 = 2 * ((precision * recall) / (precision + recall))
    else:
        precision = precision_score(y_true=y_test, y_pred=y_pred)
        recall = recall_score(y_true=y_test, y_pred=y_pred)
        f1 = f1_score(y_true=y_test, y_pred=y_pred)

    return Tp, Fp, Tn, Fn, accuracy, precision, recall, f1


def print_run_results(latex_table_print, run_num, Tp, Fp, Tn, Fn, accuracy, precision, recall, f1, duration):
    if latex_table_print:
        print("{:.0f} & 0.25 & {:.3f} & {:.3f} & {:.3f} & {:.3f} & {:.3f} & {:.3f} & {:.3f} & {:.3f} & {:.3f} ".format(
            run_num, Tp, Fp, Tn, Fn, accuracy, precision, recall, f1, duration))
        print("\hline")
    else:
        print("---Run " + str(run_num) + "---")
        print("TP: " + str(Tp))
        print("FP: " + str(Fp))
        print("TN: " + str(Tn))
        print("FN: " + str(Fn))
        print("Accuracy: " + str(accuracy))
        print("Precision: " + str(precision))
        print("Recall: " + str(recall))
        print("F1: " + str(f1))
        print("Duration: " + str(duration))

def print_average_results(latex_table_print, Tps, Fps, Tns, Fns, accs, precs, recalls, f1s, durations):
    if latex_table_print:
        print("0.25 & {:.3f} & {:.3f} & {:.3f} & {:.3f} & {:.3f} & {:.3f} & {:.3f} & {:.3f} & {:.3f} ".format(
                sum(Tps)/3, sum(Fps)/3, sum(Tns)/3, sum(Fns)/3, sum(accs)/3, sum(precs)/3, sum(recalls)/3, sum(f1s)/3, sum(durations)/3))
        print ("\hline")
    else:
        print ("---Averages---")
        print ("TP: " + str(sum(Tps)/3))
        print("FP: " + str(sum(Fns)/3))
        print("TN: " + str(sum(Tns)/3))
        print("FN: " + str(sum(Fns)/3))
        print("Accuracy: " + str(sum(accs)))
        print("Precision: " + str(sum(precs)))
        print("Recall: " + str(sum(recalls)/3))
        print("F1: " + str(sum(f1s)/3))
        print("Duration: " + str(sum(durations)/3))


def run():

    filename_features = 'values/100_perc.npy'
    filename_targets = 'values/defect_identification_below_treated_as_true.csv'
    test_size = 0.2
    k_best_features = 15500
    fraction = 0.5
    number_of_trees = 100
    number_of_runs = 3

    remove_redundant_features = True
    use_all_processor_cores = True
    use_fraction_of_dataset = False
    revised_performance_measures = True
    inspect_feature_imbalance = False
    latex_table_print = False

    x = load_feature_values(filename_features)
    y = load_target_values(filename_targets)

    if remove_redundant_features:
        x = remove_zero_variance_features(x=x)
        x = select_k_best_features(k_best_features=k_best_features, x=x, y=y)

    if inspect_feature_imbalance:
        calculate_feature_imbalance(x)

    x_train, x_test, y_train, y_test = split_dataset(x=x, y=y, test_size=test_size, random_state=666)

    if use_fraction_of_dataset:
        x_train, y_train, x_test, y_test = get_fraction_of_dataset(fraction=fraction,x_train=x_train,
                                                                   y_train=y_train, x_test=x_test, y_test=y_test)
    if not remove_redundant_features:
        x_train = x_train.reshape(x_train.shape[0], x_train.shape[1] * x_train.shape[2])
        x_test = x_test.reshape(x_test.shape[0], x_test.shape[1] * x_test.shape[2])

    if use_all_processor_cores:
        classifier = RandomForestClassifier(n_estimators=number_of_trees, n_jobs=-1)
    else:
        classifier = RandomForestClassifier(n_estimators=number_of_trees)

    Tns = []
    Tps = []
    Fns = []
    Fps = []
    accs = []
    precs = []
    recalls = []
    f1s = []
    durations = []

    for run_num in range(1, number_of_runs+1):
        time1 = time.time()
        print ("Training model...")
        fit = classifier.fit(x_train, y_train)
        print ("Model trained.")
        print ("Predicting test set...")
        y_pred = fit.predict(x_test)
        print ("Test set predicted.")
        time2 = time.time()
        print ("Evaluating...")
        duration = time2-time1

        Tp, Fp, Tn, Fn, accuracy, precision, recall, f1 = calculate_performance_scores(
                                                                                        revised_performance_measures=revised_performance_measures,
                                                                                        y_test=y_test, y_pred=y_pred)

        Tns.append(Tn)
        Tps.append(Tp)
        Fns.append(Fn)
        Fps.append(Fp)
        accs.append(accuracy)
        precs.append(precision)
        recalls.append(recall)
        f1s.append(f1)
        durations.append(duration)

        print_run_results(latex_table_print=latex_table_print, run_num=run_num, Tp=Tp, Fp=Fp, Tn=Tn, Fn=Fn,
                          accuracy=accuracy, precision=precision, recall=recall, f1=f1, duration=duration)

    print_average_results(latex_table_print=latex_table_print, Tps=Tps, Fps=Fps, Tns=Tns, Fns=Fns,
                          accs=accs, precs=precs, recalls=recalls, f1s=f1s, durations=durations)


run()



